//
//	Gordon Wetzstein [gordon.wetzstein@medien.uni-weimar.de]
//	Faculty of Media
//	Bauhaus-University Weimar
//
//-----------------------------------------------------------------------------
// includes

#include <stdio.h>
#include <iostream>
using namespace std;

#ifdef WIN32
#include <windows.h>
#endif

#include <gl/glew.h>
#include "gl/glut.h"

#include <cg/cg.h>
#include <cg/cggl.h>

#include "FramebufferObject.h"
FramebufferObject	_framebufferObject[2];
int _currentFBO = 0;



//-----------------------------------------------------------------------------

GLint	_gridSize[]	= { 64, 64 };	
bool	_bFBOInitialized = false;
float timestep = 0.0f;
int leftMouseButtonActive = 0;
GLint _windowSize[] = { 640, 480 };

//-----------------------------------------------------------------------------

/// Cg context
CGcontext	_cgContext;
///	Cg profile for the fragment shader
CGprofile	_cgFragmentProfile;
///	Cg program for testing the performance
CGprogram	_cgProgram;
CGparameter	_cgGridSize;
CGparameter	_cgHeatTexture;
CGparameter	_cgTimestep;

//pass through fragment program
CGprogram	_cgProgramPT;
CGparameter	_cgTexturePT;

bool _bShaderInitialized = false;

//-----------------------------------------------------------------------------
// GLUT callbacks and functions

void initGlut(int argc, char **argv);
void displayFunc(void);
void idleFunc(void);
void reshapeFunc(int width, int height);
void mouseFunc(int button, int state, int x, int y);
void mouseMotionFunc(int x, int y);
void keyboardFunc(unsigned char key, int x, int y);
void specialFunc(int key, int x, int y);

//-----------------------------------------------------------------------------

// other [OpenGL] functions
void countFrames(void);
void renderBitmapString(float x, float y, float z, void *font, char *string);

//-----------------------------------------------------------------------------

bool bFullsreen = false;
int nWindowID;

//-----------------------------------------------------------------------------

// parameters for the framecounter
char pixelstring[30];
int cframe = 0;
int time = 0;
int timebase = 0;

bool init = false;

bool bMousePressed = false;
int mousePressedCoord[2];

//-----------------------------------------------------------------------------
//	shader stuff

bool	bShaderInitialized = false;
void	initShader(void);
void	simulationStep(void);
void	incrementHeatAtPixel(int x, int y);

//-----------------------------------------------------------------------------

void displayFunc(void) {

	GLenum err;

	if(!_bFBOInitialized) {

		// initialize FBO
		_framebufferObject[0].initialize( _gridSize[0], _gridSize[1] );
		_framebufferObject[1].initialize( _gridSize[0], _gridSize[1] );
		_bFBOInitialized = true;

		// initialize grid with zeros
		for(int i=0; i<2; i++) {
			_framebufferObject[i].beginCapture();
			glClearColor(0.0, 0.0, 0.0, 1.0);
			glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
			_framebufferObject[i].endCapture();
		} 

		// init simulation shader too
		initShader();
		_currentFBO = 1;

		// initialize with smily
		// left eye
		incrementHeatAtPixel(_gridSize[0]/2-2,_gridSize[1]/2+3);
		incrementHeatAtPixel(_gridSize[0]/2-3,_gridSize[1]/2+3);
		incrementHeatAtPixel(_gridSize[0]/2-3,_gridSize[1]/2+2);
		incrementHeatAtPixel(_gridSize[0]/2-2,_gridSize[1]/2+2);
		// right eye
		incrementHeatAtPixel(_gridSize[0]/2+2,_gridSize[1]/2+3);
		incrementHeatAtPixel(_gridSize[0]/2+3,_gridSize[1]/2+3);
		incrementHeatAtPixel(_gridSize[0]/2+3,_gridSize[1]/2+2);
		incrementHeatAtPixel(_gridSize[0]/2+2,_gridSize[1]/2+2);
		// mouth
		incrementHeatAtPixel(_gridSize[0]/2-4,_gridSize[1]/2-1);
		incrementHeatAtPixel(_gridSize[0]/2-3,_gridSize[1]/2-2);
		incrementHeatAtPixel(_gridSize[0]/2-2,_gridSize[1]/2-3);
		incrementHeatAtPixel(_gridSize[0]/2-1,_gridSize[1]/2-4);
		incrementHeatAtPixel(_gridSize[0]/2,_gridSize[1]/2-4);
		incrementHeatAtPixel(_gridSize[0]/2+1,_gridSize[1]/2-4);
		incrementHeatAtPixel(_gridSize[0]/2+2,_gridSize[1]/2-3);
		incrementHeatAtPixel(_gridSize[0]/2+3,_gridSize[1]/2-2);
		incrementHeatAtPixel(_gridSize[0]/2+4,_gridSize[1]/2-1);
		
		// check for errors
		err = glGetError();
		if(err != GL_NO_ERROR)
			printf("OpenGL ERROR after FBO initialization: %s\n", gluErrorString(err));
	}

	// -----------------------------------------------------------------------

	// add user input
	//if(bMousePressed) {
	//	incrementHeatAtPixel(mousePressedCoord[0],mousePressedCoord[1]);
	//	bMousePressed = false;
	//}

	// advance heat equation
	simulationStep();

	// -----------------------------------------------------------------------

	// draw result
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluOrtho2D(0.0, 1.0, 0.0, 1.0);

	glMatrixMode(GL_MODELVIEW);	
	glPushMatrix();
	glLoadIdentity();

	// draw color buffer 
	_framebufferObject[_currentFBO].bind();
	glBegin(GL_QUADS);
		glTexCoord2f(0.0, 0.0);
		glVertex2f(0.0, 0.0);
		glTexCoord2f(_framebufferObject[_currentFBO].getWidth(), 0.0);
		glVertex2f(1.0, 0.0);
		glTexCoord2f(_framebufferObject[_currentFBO].getWidth(), _framebufferObject[_currentFBO].getHeight());
		glVertex2f(1.0, 1.0);
		glTexCoord2f(0.0, _framebufferObject[_currentFBO].getHeight());
		glVertex2f(0.0, 1.0);
	glEnd();

	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();

	// framecounter
	countFrames();
	
	// check for errors
	err = glGetError();
	if(err != GL_NO_ERROR)
		printf("OpenGL ERROR: %s\n", gluErrorString(err));

	glutSwapBuffers();
	_currentFBO = (_currentFBO == 0) ? 1 : 0;
}

//-----------------------------------------------------------------------------

void	simulationStep(void) {

	int _otherFBO = (_currentFBO == 0) ? 1 : 0;

	glPushAttrib(GL_ALL_ATTRIB_BITS);

	_framebufferObject[_currentFBO].beginCapture(false);

	glEnable(GL_DEPTH_TEST);
	glDisable(GL_TEXTURE_2D);
	_framebufferObject[_currentFBO].disableTextureTarget();

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluOrtho2D(0.0, 1.0, 0.0, 1.0);

	glMatrixMode(GL_MODELVIEW);	
	glPushMatrix();
	glLoadIdentity();

	// enable shader and set parameters
	cgGLEnableProfile( _cgFragmentProfile );
	cgGLBindProgram( _cgProgram );

	cgGLSetTextureParameter(_cgHeatTexture, _framebufferObject[_otherFBO].getColorAttachmentID());
	cgGLEnableTextureParameter(_cgHeatTexture);
	cgGLSetParameter2f(_cgGridSize, 1.0/_gridSize[0], 1.0/_gridSize[1]);
	cgGLSetParameter1f(_cgTimestep, 10*timestep);

	// draw quad with texture coordinates
	glBegin(GL_QUADS);
		glTexCoord2f(0.0, 0.0);
		glVertex2f(0.0, 0.0);
		glTexCoord2f(_framebufferObject[_currentFBO].getWidth(), 0.0);
		glVertex2f(1.0, 0.0);
		glTexCoord2f(_framebufferObject[_currentFBO].getWidth(), _framebufferObject[_currentFBO].getHeight());
		glVertex2f(1.0, 1.0);
		glTexCoord2f(0.0, _framebufferObject[_currentFBO].getHeight());
		glVertex2f(0.0, 1.0);
	glEnd();

	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();

	_framebufferObject[_currentFBO].endCapture(false);

	glPopAttrib();

}


//-----------------------------------------------------------------------------

void initGlut(int argc, char **argv) {

	// GLUT Window Initialization:
	glutInit (&argc, argv);
	glutInitWindowSize (_windowSize[0], _windowSize[1]);
	glutInitWindowPosition(200, 100);
	glutInitDisplayMode ( GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH | GLUT_STENCIL);
	nWindowID = glutCreateWindow ("2D Heat Equation");

	GLenum err = glewInit();
	if (GLEW_OK != err)
	{
		std::cout << "ERROR: could not initialize GLEW!" << std::endl;
	}

	// Register callbacks:
	glutDisplayFunc		(displayFunc);
	glutReshapeFunc		(reshapeFunc);
	glutKeyboardFunc	(keyboardFunc);
	glutSpecialFunc		(specialFunc);
	glutMouseFunc		(mouseFunc);
	glutMotionFunc		(mouseMotionFunc);
	glutIdleFunc		(idleFunc);

}

//-----------------------------------------------------------------------------

void	initShader(void) {

	glPushAttrib(GL_ALL_ATTRIB_BITS);

	//-------------------------------------------------------------------
	// init fragment shader profile

	_cgFragmentProfile = cgGLGetLatestProfile(CG_GL_FRAGMENT);   

	//-------------------------------------------------------------------
	// init context

	if(!_cgContext)
		_cgContext = cgCreateContext();

	//-------------------------------------------------------------------
	//	init shader

	if(_bShaderInitialized)
		return;

	_cgProgram = cgCreateProgramFromFile(	_cgContext, 
											CG_SOURCE, 
											"HeatShader.cg", 
											_cgFragmentProfile, 
											NULL, 
											NULL );

	cgGLLoadProgram( _cgProgram );
	_cgHeatTexture	= cgGetNamedParameter(_cgProgram, "heatTexture");
	_cgGridSize		= cgGetNamedParameter(_cgProgram, "invGridSize");
	_cgTimestep		= cgGetNamedParameter(_cgProgram, "timestep");

	//-------------------------------------------------------------------

	const char *passThroughProgram =
		"void main( float4 texcoord : TEXCOORD,\n"
		"			uniform samplerRECT texture,\n"
		"			out float4 color: COLOR ) {\n"
		"	color.rgba = f4texRECT(texture, texcoord.xy).rgba; }\n";

	_cgProgramPT = cgCreateProgram(	_cgContext, 
									CG_SOURCE, 
									passThroughProgram, 
									_cgFragmentProfile, 
									NULL, 
									NULL );
	cgGLLoadProgram( _cgProgramPT );
	_cgTexturePT = cgGetNamedParameter(_cgProgramPT, "texture");

	//-------------------------------------------------------------------

	GLenum error = glGetError();
	if(error != GL_NO_ERROR)
		cout << "GL ERROR: " << gluErrorString(error) << endl;

	if(CGerror cgError = cgGetError())
		cout << "Cg ERROR: " << cgGetErrorString(cgError) << endl;

	_bShaderInitialized = true;
}

//-----------------------------------------------------------------------------

void idleFunc(void) {
	glutPostRedisplay();
}

//-----------------------------------------------------------------------------

void reshapeFunc(int width, int height) {

	_windowSize[0] = width;
	_windowSize[1] = height;
	glViewport(0, 0, width, height);
}

//-----------------------------------------------------------------------------
// mouse callback
void mouseFunc(int button, int state, int x, int y) {
	

	// get the mouse buttons
	if (button == GLUT_LEFT_BUTTON) {
		if (state == GLUT_DOWN) {
			leftMouseButtonActive += 1;

			int gridX = (int)((float)x*(float)(_gridSize[0])/(float)_windowSize[0]);
			int gridY = (int)((float)(_windowSize[1]-y-1)*(float)(_gridSize[1])/(float)_windowSize[1]);
			//mousePressedCoord[0] = gridX;
			//mousePressedCoord[1] = gridY;
			//bMousePressed = true;
			incrementHeatAtPixel(gridX,gridY);
		} else {
			leftMouseButtonActive -= 1;
		}
	}
}

//-----------------------------------------------------------------------------

void mouseMotionFunc(int x, int y) {		
	if (leftMouseButtonActive) {
		int gridX = (int)((float)x*(float)(_gridSize[0])/(float)_windowSize[0]);
		int gridY = (int)((float)(_windowSize[1]-y-1)*(float)(_gridSize[1])/(float)_windowSize[1]);
		//mousePressedCoord[0] = gridX;
		//mousePressedCoord[1] = gridY;
		//bMousePressed = true;
		incrementHeatAtPixel(gridX,gridY);
	}	
}

//-----------------------------------------------------------------------------

void keyboardFunc(unsigned char key, int x, int y) {
    
	switch(key) {
		
		// -----------------------------------------

#ifdef WIN32
		// exit on escape
		case '\033':
			exit(0);
			break;
#endif

		// -----------------------------------------
			
		// switch to fullscreen
		case 'f':

			bFullsreen = !bFullsreen;
			if(bFullsreen) 
				glutFullScreen();
			else {
				glutSetWindow(nWindowID);
				glutPositionWindow(100, 100);
				glutReshapeWindow(640, 480);
			}
			break;
	}
}

//-----------------------------------------------------------------------------

void specialFunc(int key, int x, int y) {
	//printf("key pressed: %d\n", key);
}

//-----------------------------------------------------------------------------

void countFrames(void)  {

	time=glutGet(GLUT_ELAPSED_TIME);
	timestep = (time-timebase)/1000.0;
	cframe++;
	if (time - timebase > 50) {
		sprintf(pixelstring, "fps: %4.2f", cframe*1000.0/(time-timebase));		
		timebase = time;
		cframe = 0;
	// Draw status text and uni-logo:
	} 

	glDisable(GL_LIGHTING);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_TEXTURE_RECTANGLE_ARB);
	glColor3f(0.0,1.0,0.0);
	
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluOrtho2D(0, 200, 0, 200);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	// render the string
	renderBitmapString(180,5,0.0,GLUT_BITMAP_HELVETICA_10,pixelstring);
		
	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
}

//-----------------------------------------------------------------------------

void	incrementHeatAtPixel(int x, int y) {

	glPushAttrib(GL_ALL_ATTRIB_BITS);

	int _otherFBO = (_currentFBO == 0) ? 1 : 0;

	_framebufferObject[_currentFBO].beginCapture();
	glClearColor(0.0, 0.0, 0.0, 1.0);
	glClear(GL_COLOR_BUFFER_BIT);

	// make a single heat point
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_TEXTURE_2D);

	glMatrixMode(GL_MODELVIEW);	
	glPushMatrix();
	glLoadIdentity();

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluOrtho2D(0.0, 1.0, 0.0, 1.0);

	// draw color buffer 
	cgGLEnableProfile( _cgFragmentProfile );
	cgGLBindProgram( _cgProgramPT );
	cgGLSetTextureParameter(_cgTexturePT, _framebufferObject[_otherFBO].getColorAttachmentID());
	cgGLEnableTextureParameter(_cgTexturePT);

	_framebufferObject[_otherFBO].bind();
	glBegin(GL_QUADS);
		glTexCoord2f(0.0, 0.0);
		glVertex2f(0.0, 0.0);
		glTexCoord2f(_framebufferObject[_otherFBO].getWidth(), 0.0);
		glVertex2f(1.0, 0.0);
		glTexCoord2f(_framebufferObject[_otherFBO].getWidth(), _framebufferObject[_otherFBO].getHeight());
		glVertex2f(1.0, 1.0);
		glTexCoord2f(0.0, _framebufferObject[_otherFBO].getHeight());
		glVertex2f(0.0, 1.0);
	glEnd();

	cgGLDisableProfile( _cgFragmentProfile );

	//------------------------------------------------------------------------------
	//------------------------------------------------------------------------------
	// enable OpenGL blending

	glEnable(GL_BLEND);
	//glBlendEquation(GL_FUNC_ADD);
	glBlendFunc(GL_ONE,GL_ONE);

	_framebufferObject[_currentFBO].disableTextureTarget();

	// draw the pixel and add to FBO
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluOrtho2D(0.0, _gridSize[0], 0.0, _gridSize[1]);

	glColor3f(1.0, 0.0, 0.0);
	glBegin(GL_QUADS);
		glVertex2f((float)x, (float)y);
		glVertex2f((float)x+1, (float)y);
		glVertex2f((float)x+1, (float)y+1);
		glVertex2f((float)x, (float)y+1);
	glEnd();

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	_framebufferObject[_currentFBO].endCapture();

//	glDisable(GL_BLEND);

	_currentFBO = _otherFBO;

	glPopAttrib();
}

//-----------------------------------------------------------------------------

void renderBitmapString(float x, float y, float z, void *font, char *string) {
  char *c;
  glRasterPos3f(x, y,z);
  for (c=string; *c != '\0'; c++) {
    glutBitmapCharacter(font, *c);
  }
}

//-----------------------------------------------------------------------------


void main(int argc, char **argv) {
	
	initGlut(argc, argv);

	printf("OpenGL version: %s\n", glGetString(GL_VERSION));
	printf("OpenGL vendor: %s\n", glGetString(GL_VENDOR));
	printf("OpenGL renderer: %s\n\n", glGetString(GL_RENDERER));
	
	printf("The Heat Equation in 2D on the GPU\n");
	printf("discretized using centered finite differences in space and forward Euler in time\n");
	printf("Gordon Wetzstein [wetzste1@cs.ubc.ca]\n\n");

	printf("click left mouse button and move to apply heat!\n");

	glutMainLoop();
		
}



